unit quirc;
(* quirc -- QR-code recognition library
 * Copyright (C) 2010-2012 Daniel Beer <dlbeer@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)

(* C to Pascal translation by Ladislav Karrach in 2018 (with help of c2pas32)
 *  Usage:
 *    q := quirc_new();
 *    if quirc_resize(q, Width, Height) = 0 then begin
 *      qimage := quirc_begin(q, nil, nil);
 *      // Fill out the image buffer here. qimage is a pointer to a Width*Height bytes.
 *      // One byte per pixel, Width pixels per line, Height lines in the buffer.
 *      quirc_end(q);
 *      for i:=0 to quirc_count(q)-1 do begin
 *        quirc_extract(q, i, qcode);
 *        if quirc_decode(@qcode, qdata) = QUIRC_SUCCESS then
 *          SetString(s, @qdata.payload, qdata.payload_len);
 *      end;
 *    end;
 *    quirc_destroy(qreader);
 *)

{$SMARTLINK ON}

interface

(* quirc.h *)

type
  int = integer;
  uint8_t = byte;
  uint16_t = word;
  uint32_t = dword;

  puint8_t = ^uint8_t;
  puint16_t = ^uint16_t;

  (* This structure describes a location in the input image buffer. *)
  quirc_point = record
	  x,y: int;
  end;
  pquirc_point = ^quirc_point;

  (* This enum describes the various decoder errors which may occur. *)
  quirc_decode_error_t = (
	  QUIRC_SUCCESS = 0,
	  QUIRC_ERROR_INVALID_GRID_SIZE,
	  QUIRC_ERROR_INVALID_VERSION,
	  QUIRC_ERROR_FORMAT_ECC,
	  QUIRC_ERROR_DATA_ECC,
	  QUIRC_ERROR_UNKNOWN_DATA_TYPE,
	  QUIRC_ERROR_DATA_OVERFLOW,
	  QUIRC_ERROR_DATA_UNDERFLOW
  );

const
  (* Limits on the maximum size of QR-codes and their content. *)
  QUIRC_MAX_BITMAP = 3917;
  QUIRC_MAX_PAYLOAD = 8896;

  (* QR-code ECC types. *)
  QUIRC_ECC_LEVEL_M = 0;
  QUIRC_ECC_LEVEL_L = 1;
  QUIRC_ECC_LEVEL_H = 2;
  QUIRC_ECC_LEVEL_Q = 3;

  (* QR-code data types. *)
  QUIRC_DATA_TYPE_NUMERIC = 1;
  QUIRC_DATA_TYPE_ALPHA   = 2;
  QUIRC_DATA_TYPE_BYTE    = 4;
  QUIRC_DATA_TYPE_KANJI   = 8;

  (* Common character encodings *)
  QUIRC_ECI_ISO_8859_1  = 1;
  QUIRC_ECI_IBM437      = 2;
  QUIRC_ECI_ISO_8859_2  = 4;
  QUIRC_ECI_ISO_8859_3  = 5;
  QUIRC_ECI_ISO_8859_4  = 6;
  QUIRC_ECI_ISO_8859_5  = 7;
  QUIRC_ECI_ISO_8859_6  = 8;
  QUIRC_ECI_ISO_8859_7  = 9;
  QUIRC_ECI_ISO_8859_8  = 10;
  QUIRC_ECI_ISO_8859_9  = 11;
  QUIRC_ECI_WINDOWS_874 = 13;
  QUIRC_ECI_ISO_8859_13 = 15;
  QUIRC_ECI_ISO_8859_15 = 17;
  QUIRC_ECI_SHIFT_JIS   = 20;
  QUIRC_ECI_UTF_8       = 26;

type
  (* This structure is used to return information about detected QR codes
   * in the input image.
   *)
  quirc_code = record
  	(* The four corners of the QR-code, from top left, clockwise *)
  	corners: array[0..3] of quirc_point;

  	(* The number of cells across in the QR-code. The cell bitmap
  	 * is a bitmask giving the actual values of cells. If the cell
  	 * at (x, y) is black, then the following bit is set:
  	 *
  	 *     cell_bitmap[i >> 3] & (1 << (i & 7))
  	 *
  	 * where i = (y * size) + x.
  	 *)
  	size: int;
  	cell_bitmap: array[0..QUIRC_MAX_BITMAP-1] of uint8_t;
  end;
  pquirc_code = ^quirc_code;

  (* This structure holds the decoded QR-code data *)
  quirc_data = record
  	(* Various parameters of the QR-code. These can mostly be
  	 * ignored if you only care about the data.
  	 *)
  	version: int;
  	ecc_level: int;
  	mask: int;

  	(* This field is the highest-valued data type found in the QR
  	 * code.
  	 *)
  	data_type: int;

  	(* Data payload. For the Kanji datatype, payload is encoded as
  	 * Shift-JIS. For all other datatypes, payload is ASCII text.
  	 *)
  	payload: array [0..QUIRC_MAX_PAYLOAD-1] of uint8_t;
  	payload_len: int;

  	(* ECI assignment number *)
  	eci: uint32_t;
  end;
  pquirc_data = ^quirc_data;


(* quirc_internal.h *)

const
  QUIRC_PIXEL_WHITE	= 0;
  QUIRC_PIXEL_BLACK	= 1;
  QUIRC_PIXEL_REGION=	2;

  QUIRC_MAX_REGIONS = 65534;  // 254 or 65534
  QUIRC_MAX_CAPSTONES = 4*52; // 32
  QUIRC_MAX_GRIDS = 52;       // 8

  QUIRC_PERSPECTIVE_PARAMS = 8;

type
  {$IF QUIRC_MAX_REGIONS < 255}
  quirc_pixel_t = uint8_t;
  {$ELSEIF QUIRC_MAX_REGIONS < 65535}
  quirc_pixel_t = uint16_t;
  {$ELSE}
  {$ERROR QUIRC_MAX_REGIONS > 65534 is not supported!}
  {$ENDIF}
  pquirc_pixel_t = ^quirc_pixel_t;

  quirc_region = record
  	seed: quirc_point;
  	count: int;
  	capstone: int;
  end;
  pquirc_region = ^quirc_region;

  quirc_capstone = record
  	ring: int;
  	stone: int;

  	corners: array[0..3] of quirc_point;
  	center: quirc_point;
  	c: array [0..QUIRC_PERSPECTIVE_PARAMS-1] of double;

  	qr_grid: int;
  end;
  pquirc_capstone = ^quirc_capstone;

  quirc_grid = record
  	(* Capstone indices *)
  	caps: array[0..2] of int;

  	(* Alignment pattern region and corner *)
  	align_region: int;
  	align: quirc_point;

  	(* Timing pattern endpoints *)
  	tpep: array[0..2] of quirc_point;
  	hscan: int;
  	vscan: int;

  	(* Grid size and perspective transform *)
  	grid_size: int;
  	c: array[0..QUIRC_PERSPECTIVE_PARAMS-1] of double;
  end;
  pquirc_grid = ^quirc_grid;

  tquirc = record
  	image: puint8_t;
  	pixels: pquirc_pixel_t;
  	row_average: ^int; (* used by threshold() *)
  	w,h: int;

  	num_regions: int;
  	regions: array[0..QUIRC_MAX_REGIONS-1] of quirc_region;

  	num_capstones: int;
  	capstones: array[0..QUIRC_MAX_CAPSTONES-1] of quirc_capstone;

  	num_grids: int;
  	grids: array[0..QUIRC_MAX_GRIDS-1] of quirc_grid;
  end;
  pquirc = ^tquirc;

(************************************************************************
 * QR-code version information database
 *)
const
  QUIRC_MAX_VERSION = 40;
  QUIRC_MAX_ALIGNMENT = 7;

type
  quirc_rs_params = record
	  bs: int; (* Small block size *)
    dw: int; (* Small data words *)
	  ns: int; (* Number of small blocks *)
  end;
  pquirc_rs_params = ^quirc_rs_params;

  quirc_version_info = record
  	data_bytes: int;
  	apat: array[0..QUIRC_MAX_ALIGNMENT-1] of int;
  	ecc: array[0..3] of quirc_rs_params;
  end;
  pquirc_version_info = ^quirc_version_info;


(* quirc.c *)

(* Obtain the library version string. *)
function quirc_version(): PAnsiChar;

(* Construct a new QR-code recognizer. This function will return NULL
 * if sufficient memory could not be allocated.
 *)
function quirc_new(): pquirc;

(* Destroy a QR-code recognizer. *)
procedure quirc_destroy(q: pquirc);

(* Resize the QR-code recognizer. The size of an image must be
 * specified before codes can be analyzed.
 *
 * This function returns 0 on success, or -1 if sufficient memory could
 * not be allocated.
 *)
function quirc_resize(q: pquirc; w,h: int): int;

(* Return the number of QR-codes identified in the last processed
 * image.
 *)
function quirc_count(q: pquirc): integer;


implementation

function quirc_version(): PAnsiChar;
begin
  Result := '1.0';
end;

function quirc_new(): pquirc;
begin
  Result := AllocMem(SizeOf(pquirc^));
end;

procedure quirc_destroy(q: pquirc);
begin
  FreeMem(q^.image);
  (* q->pixels may alias q->image when their type representation is of the
      same size, so we need to be careful here to avoid a double free *)
  if sizeof(q^.image^) <> sizeof(q^.pixels^) then
    FreeMem(q^.pixels);
  FreeMem(q^.row_average);
  FreeMem(q);
end;

function quirc_resize(q: pquirc; w,h: int): int;
var
	image: puint8_t;
	pixels: pquirc_pixel_t;
	row_average: pinteger;
  olddim, newdim, min: integer;
label
  fail;
begin
  image := nil;
  pixels := nil;
  row_average := nil;

	(*
	 * XXX: w and h should be size_t (or at least unsigned) as negatives
	 * values would not make much sense. The downside is that it would break
	 * both the API and ABI. Thus, at the moment, let's just do a sanity
	 * check.
	 *)
	if (w < 0) or (h < 0) then goto fail;

	(*
	 * alloc a new buffer for q->image. We avoid realloc(3) because we want
	 * on failure to be leave `q` in a consistant, unmodified state.
	 *)
	image := AllocMem(w*h);
	if image=nil then	goto fail;

	(* compute the "old" (i.e. currently allocated) and the "new"
	   (i.e. requested) image dimensions *)
	olddim := q^.w * q^.h;
	newdim := w * h;
	if (olddim < newdim) then min := olddim else min := newdim;

	(*
	 * copy the data into the new buffer, avoiding (a) to read beyond the
	 * old buffer when the new size is greater and (b) to write beyond the
	 * new buffer when the new size is smaller, hence the min computation.
	 *)
	Move(q^.image^, image^, min);

	(* alloc a new buffer for q->pixels if needed *)
	if sizeof(q^.image^) <> sizeof(q^.pixels^) then begin
		pixels := AllocMem(newdim * sizeof(quirc_pixel_t));
		if pixels = nil then goto fail;
  end;

	(* alloc a new buffer for q->row_average *)
	row_average := AllocMem(w * sizeof(int));
	if row_average = nil then goto fail;

	(* alloc succeeded, update `q` with the new size and buffers *)
	q^.w := w;
	q^.h := h;
	FreeMem(q^.image);
	q^.image := image;
	if sizeof(q^.image^) <> sizeof(q^.pixels^) then begin
		FreeMem(q^.pixels);
		q^.pixels := pixels;
	end;
  FreeMem(q^.row_average);
	q^.row_average := row_average;

	Exit(0);
	(* NOTREACHED *)
fail:
	FreeMem(image);
	FreeMem(pixels);
	FreeMem(row_average);

	Exit(-1);
end;

function quirc_count(q: pquirc): integer;
begin
  Result := q^.num_grids;
end;

end.
