unit quircdecode;
(* quirc -- QR-code recognition library
 * Copyright (C) 2010-2012 Daniel Beer <dlbeer@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)

(* C to Pascal translation by Ladislav Karrach in 2018 (with help of c2pas32)
 *  Usage:
 *    // QR Code matrix must be encoded in code.cell_bitmap and code.size must be set
 *    i := 0;
 *    for y:=0 to Image.Height - 1 do
 *      for x:=0 to Image.Width - 1 do begin
 *        if Image.Pixels[x,y] is black module then
 *          code.cell_bitmap[i shr 3] := code.cell_bitmap[i shr 3] or (1 shl (i and 7));
 *        Inc(i);
 *      end;
 *    if quirc_decode(@code, data) = QUIRC_SUCCESS then ...
 *)

interface

uses quirc;

(* Decode a QR-code, returning the payload data. *)
function quirc_decode(code: pquirc_code; out data: quirc_data): quirc_decode_error_t;

const
  (* version_db.c *)
  quirc_version_db: array [0..QUIRC_MAX_VERSION] of quirc_version_info = (
    (data_bytes: 0; apat:(0,0,0,0,0,0,0); ecc:((bs:0;dw:0;ns:0),(bs:0;dw:0;ns:0),(bs:0;dw:0;ns:0),(bs:0;dw:0;ns:0))),
    // Version 1
    (data_bytes:26; apat:(0,0,0,0,0,0,0); ecc:((bs:26;dw:16;ns:1),(bs:26;dw:19;ns:1),(bs:26;dw:9;ns:1),(bs:26;dw:13;ns:1))),
    // Version 2
    (data_bytes:44; apat:(6,18,0,0,0,0,0); ecc:((bs:44;dw:28;ns:1),(bs:44;dw:34;ns:1),(bs:44;dw:16;ns:1),(bs:44;dw:22;ns:1))),
    // Version 3
    (data_bytes:70; apat:(6,22,0,0,0,0,0); ecc:((bs:70;dw:44;ns:1),(bs:70;dw:55;ns:1),(bs:35;dw:13;ns:2),(bs:35;dw:17;ns:2))),
    // Version 4
    (data_bytes:100; apat:(6,26,0,0,0,0,0); ecc:((bs:50;dw:32;ns:2),(bs:100;dw:80;ns:1),(bs:25;dw:9;ns:4),(bs:50;dw:24;ns:2))),
    // Version 5
    (data_bytes:134; apat:(6,30,0,0,0,0,0); ecc:((bs:67;dw:43;ns:2),(bs:134;dw:108;ns:1),(bs:33;dw:11;ns:2),(bs:33;dw:15;ns:2))),
    // Version 6
    (data_bytes:172; apat:(6,34,0,0,0,0,0); ecc:((bs:43;dw:27;ns:4),(bs:86;dw:68;ns:2),(bs:43;dw:15;ns:4),(bs:43;dw:19;ns:4))),
    // Version 7
    (data_bytes:196; apat:(6,22,38,0,0,0,0); ecc:((bs:49;dw:31;ns:4),(bs:98;dw:78;ns:2),(bs:39;dw:13;ns:4),(bs:32;dw:14;ns:2))),
    // Version 8
    (data_bytes:242; apat:(6,24,42,0,0,0,0); ecc:((bs:60;dw:38;ns:2),(bs:121;dw:97;ns:2),(bs:40;dw:14;ns:4),(bs:40;dw:18;ns:4))),
    // Version 9
    (data_bytes:292; apat:(6,26,46,0,0,0,0); ecc:((bs:58;dw:36;ns:3),(bs:146;dw:116;ns:2),(bs:36;dw:12;ns:4),(bs:36;dw:16;ns:4))),
    // Version 10
    (data_bytes:346; apat:(6,28,50,0,0,0,0); ecc:((bs:69;dw:43;ns:4),(bs:86;dw:68;ns:2),(bs:43;dw:15;ns:6),(bs:43;dw:19;ns:6))),
    // Version 11
    (data_bytes:404; apat:(6,30,54,0,0,0,0); ecc:((bs:80;dw:50;ns:1),(bs:101;dw:81;ns:4),(bs:36;dw:12;ns:3),(bs:50;dw:22;ns:4))),
    // Version 12
    (data_bytes:466; apat:(6,32,58,0,0,0,0); ecc:((bs:58;dw:36;ns:6),(bs:116;dw:92;ns:2),(bs:42;dw:14;ns:7),(bs:46;dw:20;ns:4))),
    // Version 13
    (data_bytes:532; apat:(6,34,62,0,0,0,0); ecc:((bs:59;dw:37;ns:8),(bs:133;dw:107;ns:4),(bs:33;dw:11;ns:12),(bs:44;dw:20;ns:8))),
    // Version 14
    (data_bytes:581; apat:(6,26,46,66,0,0,0); ecc:((bs:64;dw:40;ns:4),(bs:145;dw:115;ns:3),(bs:36;dw:12;ns:11),(bs:36;dw:16;ns:11))),
    // Version 15
    (data_bytes:655; apat:(6,26,48,70,0,0,0); ecc:((bs:65;dw:41;ns:5),(bs:109;dw:87;ns:5),(bs:36;dw:12;ns:11),(bs:54;dw:24;ns:5))),
    // Version 16
    (data_bytes:733; apat:(6,26,50,74,0,0,0); ecc:((bs:73;dw:45;ns:7),(bs:122;dw:98;ns:5),(bs:45;dw:15;ns:3),(bs:43;dw:19;ns:15))),
    // Version 17
    (data_bytes:815; apat:(6,30,54,78,0,0,0); ecc:((bs:74;dw:46;ns:10),(bs:135;dw:107;ns:1),(bs:42;dw:14;ns:2),(bs:50;dw:22;ns:1))),
    // Version 18
    (data_bytes:901; apat:(6,30,56,82,0,0,0); ecc:((bs:69;dw:43;ns:9),(bs:150;dw:120;ns:5),(bs:42;dw:14;ns:2),(bs:50;dw:22;ns:17))),
    // Version 19
    (data_bytes:991; apat:(6,30,58,86,0,0,0); ecc:((bs:70;dw:44;ns:3),(bs:141;dw:113;ns:3),(bs:39;dw:13;ns:9),(bs:47;dw:21;ns:17))),
    // Version 20
    (data_bytes:1085; apat:(6,34,62,90,0,0,0); ecc:((bs:67;dw:41;ns:3),(bs:135;dw:107;ns:3),(bs:43;dw:15;ns:15),(bs:54;dw:24;ns:15))),
    // Version 21
    (data_bytes:1156; apat:(6,28,50,72,92,0,0); ecc:((bs:68;dw:42;ns:17),(bs:144;dw:116;ns:4),(bs:46;dw:16;ns:19),(bs:50;dw:22;ns:17))),
    // Version 22
    (data_bytes:1258;apat:(6,26,50,74,98,0,0);ecc:((bs:74;dw:46;ns:17),(bs:139;dw:111;ns:2),(bs:37;dw:13;ns:34),(bs:54;dw:24;ns:7))),
    // Version 23
    (data_bytes:1364;apat:(6,30,54,78,102,0,0);ecc:((bs:75;dw:47;ns:4),(bs:151;dw:121;ns:4),(bs:45;dw:15;ns:16),(bs:54;dw:24;ns:11))),
    // Version 24
    (data_bytes:1474;apat:(6,28,54,80,106,0,0);ecc:((bs:73;dw:45;ns:6),(bs:147;dw:117;ns:6),(bs:46;dw:16;ns:30),(bs:54;dw:24;ns:11))),
    // Version 25
    (data_bytes:1588;apat:(6,32,58,84,110,0,0);ecc:((bs:75;dw:47;ns:8),(bs:132;dw:106;ns:8),(bs:45;dw:15;ns:22),(bs:54;dw:24;ns:7))),
    // Version 26
    (data_bytes:1706;apat:(6,30,58,86,114,0,0);ecc:((bs:74;dw:46;ns:19),(bs:142;dw:114;ns:10),(bs:46;dw:16;ns:33),(bs:50;dw:22;ns:28))),
    // Version 27
    (data_bytes:1828;apat:(6,34,62,90,118,0,0);ecc:((bs:73;dw:45;ns:22),(bs:152;dw:122;ns:8),(bs:45;dw:15;ns:12),(bs:53;dw:23;ns:8))),
    // Version 28
    (data_bytes:1921;apat:(6,26,50,74,98,122,0);ecc:((bs:73;dw:45;ns:3),(bs:147;dw:117;ns:3),(bs:45;dw:15;ns:11),(bs:54;dw:24;ns:4))),
    // Version 29
    (data_bytes:2051;apat:(6,30,54,78,102,126,0);ecc:((bs:73;dw:45;ns:21),(bs:146;dw:116;ns:7),(bs:45;dw:15;ns:19),(bs:53;dw:23;ns:1))),
    // Version 30
    (data_bytes:2185;apat:(6,26,52,78,104,130,0);ecc:((bs:75;dw:47;ns:19),(bs:145;dw:115;ns:5),(bs:45;dw:15;ns:23),(bs:54;dw:24;ns:15))),
    // Version 31
    (data_bytes:2323;apat:(6,30,56,82,108,134,0);ecc:((bs:74;dw:46;ns:2),(bs:145;dw:115;ns:13),(bs:45;dw:15;ns:23),(bs:54;dw:24;ns:42))),
    // Version 32
    (data_bytes:2465;apat:(6,34,60,86,112,138,0);ecc:((bs:74;dw:46;ns:10),(bs:145;dw:115;ns:17),(bs:45;dw:15;ns:19),(bs:54;dw:24;ns:10))),
    // Version 33
    (data_bytes:2611;apat:(6,30,58,86,114,142,0);ecc:((bs:74;dw:46;ns:14),(bs:145;dw:115;ns:17),(bs:45;dw:15;ns:11),(bs:54;dw:24;ns:29))),
    // Version 34
    (data_bytes:2761;apat:(6,34,62,90,118,146,0);ecc:((bs:74;dw:46;ns:14),(bs:145;dw:115;ns:13),(bs:46;dw:16;ns:59),(bs:54;dw:24;ns:44))),
    // Version 35
    (data_bytes:2876;apat:(6,30,54,78,102,126,150);ecc:((bs:75;dw:47;ns:12),(bs:151;dw:121;ns:12),(bs:45;dw:15;ns:22),(bs:54;dw:24;ns:39))),
    // Version 36
    (data_bytes:3034;apat:(6,24,50,76,102,128,154);ecc:((bs:75;dw:47;ns:6),(bs:151;dw:121;ns:6),(bs:45;dw:15;ns:2),(bs:54;dw:24;ns:46))),
    // Version 37
    (data_bytes:3196;apat:(6,28,54,80,106,132,158);ecc:((bs:74;dw:46;ns:29),(bs:152;dw:122;ns:17),(bs:45;dw:15;ns:24),(bs:54;dw:24;ns:49))),
    // Version 38
    (data_bytes:3362;apat:(6,32,58,84,110,136,162);ecc:((bs:74;dw:46;ns:13),(bs:152;dw:122;ns:4),(bs:45;dw:15;ns:42),(bs:54;dw:24;ns:48))),
    // Version 39
    (data_bytes:3532;apat:(6,26,54,82,110,138,166);ecc:((bs:75;dw:47;ns:40),(bs:147;dw:117;ns:20),(bs:45;dw:15;ns:10),(bs:54;dw:24;ns:43))),
    // Version 40
    (data_bytes:3706;apat:(6,30,58,86,114,142,170);ecc:((bs:75;dw:47;ns:18),(bs:148;dw:118;ns:19),(bs:45;dw:15;ns:20),(bs:54;dw:24;ns:34)))
  );


implementation

type
  galois_field = record
    p: int;
    log: puint8_t;
    exp: puint8_t;
  end;
  pgalois_field = ^galois_field;

const
  MAX_POLY = 64;

  (************************************************************************
   * Galois fields
   *)
  gf16_exp: array [0..15] of uint8_t = ($01,$02,$04,$08,$03,$06,$0c,$0b,$05,$0a,$07,$0e,$0f,$0d,$09,$01);
  gf16_log: array [0..15] of uint8_t = ($00,$0f,$01,$04,$02,$08,$05,$0a,$03,$0e,$09,$07,$06,$0d,$0b,$0c);
  gf16: galois_field = (p:15; log:@gf16_log; exp:@gf16_exp);
  gf256_exp: array [0..255] of uint8_t = ($01,$02,$04,$08,$10,$20,$40,$80,$1d,$3a,$74,$e8,$cd,$87,$13,$26,$4c,$98,$2d,$5a,$b4,$75,$ea,$c9,$8f,$03,$06,$0c,$18,$30,$60,$c0,$9d,$27,$4e,$9c,$25,$4a,$94,$35,$6a,$d4,$b5,$77,$ee,$c1,$9f,$23,$46,$8c,$05,$0a,$14,$28,$50,$a0,$5d,$ba,$69,$d2,$b9,$6f,$de,$a1,$5f,$be,$61,$c2,$99,$2f,$5e,$bc,$65,$ca,$89,$0f,$1e,$3c,$78,$f0,$fd,$e7,$d3,$bb,$6b,$d6,$b1,$7f,$fe,$e1,$df,$a3,$5b,$b6,$71,$e2,$d9,$af,$43,$86,$11,$22,$44,$88,$0d,$1a,$34,$68,$d0,$bd,$67,$ce,$81,$1f,$3e,$7c,$f8,$ed,$c7,$93,$3b,$76,$ec,$c5,$97,$33,$66,$cc,$85,$17,$2e,$5c,$b8,$6d,$da,$a9,$4f,$9e,$21,$42,$84,$15,$2a,$54,$a8,$4d,$9a,$29,$52,$a4,$55,$aa,$49,$92,$39,$72,$e4,$d5,$b7,$73,$e6,$d1,$bf,$63,$c6,$91,$3f,$7e,$fc,$e5,$d7,$b3,$7b,$f6,$f1,$ff,$e3,$db,$ab,$4b,$96,$31,$62,$c4,$95,$37,$6e,$dc,$a5,$57,$ae,$41,$82,$19,$32,$64,$c8,$8d,$07,$0e,$1c,$38,$70,$e0,$dd,$a7,$53,$a6,$51,$a2,$59,$b2,$79,$f2,$f9,$ef,$c3,$9b,$2b,$56,$ac,$45,$8a,$09,$12,$24,$48,$90,$3d,$7a,$f4,$f5,$f7,$f3,$fb,$eb,$cb,$8b,$0b,$16,$2c,$58,$b0,$7d,$fa,$e9,$cf,$83,$1b,$36,$6c,$d8,$ad,$47,$8e,$01);
  gf256_log: array [0..255] of uint8_t = ($00,$ff,$01,$19,$02,$32,$1a,$c6,$03,$df,$33,$ee,$1b,$68,$c7,$4b,$04,$64,$e0,$0e,$34,$8d,$ef,$81,$1c,$c1,$69,$f8,$c8,$08,$4c,$71,$05,$8a,$65,$2f,$e1,$24,$0f,$21,$35,$93,$8e,$da,$f0,$12,$82,$45,$1d,$b5,$c2,$7d,$6a,$27,$f9,$b9,$c9,$9a,$09,$78,$4d,$e4,$72,$a6,$06,$bf,$8b,$62,$66,$dd,$30,$fd,$e2,$98,$25,$b3,$10,$91,$22,$88,$36,$d0,$94,$ce,$8f,$96,$db,$bd,$f1,$d2,$13,$5c,$83,$38,$46,$40,$1e,$42,$b6,$a3,$c3,$48,$7e,$6e,$6b,$3a,$28,$54,$fa,$85,$ba,$3d,$ca,$5e,$9b,$9f,$0a,$15,$79,$2b,$4e,$d4,$e5,$ac,$73,$f3,$a7,$57,$07,$70,$c0,$f7,$8c,$80,$63,$0d,$67,$4a,$de,$ed,$31,$c5,$fe,$18,$e3,$a5,$99,$77,$26,$b8,$b4,$7c,$11,$44,$92,$d9,$23,$20,$89,$2e,$37,$3f,$d1,$5b,$95,$bc,$cf,$cd,$90,$87,$97,$b2,$dc,$fc,$be,$61,$f2,$56,$d3,$ab,$14,$2a,$5d,$9e,$84,$3c,$39,$53,$47,$6d,$41,$a2,$1f,$2d,$43,$d8,$b7,$7b,$a4,$76,$c4,$17,$49,$ec,$7f,$0c,$6f,$f6,$6c,$a1,$3b,$52,$29,$9d,$55,$aa,$fb,$60,$86,$b1,$bb,$cc,$3e,$5a,$cb,$59,$5f,$b0,$9c,$a9,$a0,$51,$0b,$f5,$16,$eb,$7a,$75,$2c,$d7,$4f,$ae,$d5,$e9,$e6,$e7,$ad,$e8,$74,$d6,$f4,$ea,$a8,$50,$58,$af);
  gf256: galois_field = (p:255; log:@gf256_log; exp:@gf256_exp);


(************************************************************************
 * Polynomial operations
 *)
procedure poly_add(dst: puint8_t;  src: puint8_t;  c: uint8_t;  shift: integer;  gf: pgalois_field);
var
  i: integer;
  log_c: integer;
  p: integer;
  v: uint8_t;
begin
  log_c := gf^.log[c];
  if c=0 then exit;
  for i:=0 to Pred(MAX_POLY) do begin
    p := i+shift;
    v := src[i];
    if (p<0) or (p>=MAX_POLY) then continue;
    if v=0 then continue;
    dst[p] := dst[p] xor (gf^.exp[(gf^.log[v]+log_c) mod gf^.p]);
  end;
end;

function poly_eval(s: puint8_t;  x: uint8_t;  gf: pgalois_field): uint8_t; 
var
  i: integer;
  log_x: uint8_t;
  c: uint8_t;
begin
  if x = 0 then
    Result := s[0]
  else begin
    Result := 0;
    log_x := gf^.log[x];
    for i:=0 to Pred(MAX_POLY) do begin
      c:=s[i];
      if c = 0 then continue;
      Result := Result xor (gf^.exp[(gf^.log[c]+log_x*i) mod gf^.p]);
    end;
  end;
end;

(************************************************************************
 * Berlekamp-Massey algorithm for finding error locator polynomials.
 *)
procedure berlekamp_massey(s: puint8_t;  N: integer;  gf: pgalois_field;  sigma: puint8_t);
var
  C: array [0..Pred(MAX_POLY)] of uint8_t;
  B: array [0..Pred(MAX_POLY)] of uint8_t;
  L: integer;

  m: integer;
  bb: uint8_t;
  d: uint8_t;
 
  mult: uint8_t;
  j,i: integer;
  T: array [0..Pred(MAX_POLY)] of uint8_t;
begin
  L:=0;
  m:=1; 
  bb:=1; 
  
  FillByte(B, sizeof(B), 0);
  FillByte(C, sizeof(C), 0);
  B[0]:= 1; 
  C[0]:= 1; 
  for j:=0 to Pred(N) do begin
    d:=s[j];
    for i:=1 to L do begin
      if (C[i]=0) or (s[j-i]=0) then continue;
      d := d xor (gf^.exp[(gf^.log[C[i]] + gf^.log[s[j-i]]) mod gf^.p]);
    end;

    mult := gf^.exp[(gf^.p - gf^.log[bb] + gf^.log[d]) mod gf^.p];
    if d = 0 then
      Inc(m)
    else if L*2 <= j then begin
      Move(C,T,sizeof(T));
      poly_add(C,B,mult,m,gf); 
      Move(T,B,sizeof(B));
      L:= j+1-L;
      bb:= d; 
      m:= 1; 
    end
    else begin
      poly_add(C,B,mult,m,gf); 
      Inc(m);
    end;
  end;
  Move(C, sigma^, MAX_POLY);
end;

(************************************************************************
 * Code stream error correction
 *
 * Generator polynomial for GF(2^8) is x^8 + x^4 + x^3 + x^2 + 1
 *)
function block_syndromes(data: puint8_t;  bs: integer;  npar: integer;  s: puint8_t): boolean;
var
  i,j: integer;
  c: uint8_t;
begin
  Result := False;
  FillByte(s^, MAX_POLY, 0);

  for i:=0 to Pred(npar) do begin
    for j:=0 to Pred(bs) do begin
      c:=data[bs-j-1]; 
      if c = 0 then continue;
      s[i] := s[i] xor gf256_exp[(int(gf256_log[c])+i*j) mod 255];
    end;
    if s[i]<>0 then Result := True;
  end;
end;

procedure eloc_poly(omega, s, sigma: puint8_t;  npar: integer);
var
  i,j: integer;
  a,b, log_a: uint8_t;
begin
  FillByte(omega^, MAX_POLY, 0);
  for i:=0 to Pred(npar) do begin
    a := sigma[i];
    log_a := gf256_log[a];

    if a = 0 then continue;

    j := 0;
    while j+1 < MAX_POLY do begin
      b := s[j+1];

      if i+j >= npar then break;

      if b <> 0 then
        omega[i+j] := omega[i+j] xor gf256_exp[(log_a+gf256_log[b]) mod 255];

      Inc(j);
    end;
  end;
end;

function correct_block(data: puint8_t;  ecc: pquirc_rs_params): quirc_decode_error_t; 
var
  npar, i: integer;
  s, sigma, sigma_deriv, omega: array [0..Pred(MAX_POLY)] of uint8_t;
  xinv, sd_x, omega_x, error: uint8_t;
begin
  npar := ecc^.bs - ecc^.dw;
  
  (* Compute syndrome vector *)
  if not block_syndromes(data, ecc^.bs, npar, s) then begin
    Result:= QUIRC_SUCCESS;
    Exit;
  end;

  berlekamp_massey(s, npar, @gf256, sigma);

  (* Compute derivative of sigma *)
  FillByte(sigma_deriv, MAX_POLY, 0);
  i := 0;
  while i+1 < MAX_POLY do begin
    sigma_deriv[i] := sigma[i+1];
    Inc(i, 2);
  end;

  (* Compute error evaluator polynomial *)
  eloc_poly(omega,s,sigma,npar-1);

  (* Find error locations and magnitudes *)
  for i:=0 to Pred(ecc^.bs) do begin
    xinv := gf256_exp[255-i];
    if poly_eval(sigma,xinv,@gf256) = 0 then begin
      sd_x := poly_eval(sigma_deriv, xinv, @gf256);
      omega_x := poly_eval(omega, xinv, @gf256);
      error := gf256_exp[(255-gf256_log[sd_x]+gf256_log[omega_x]) mod 255];
      data[ecc^.bs-i-1] := data[ecc^.bs-i-1] xor error;
    end;
  end;

  if block_syndromes(data, ecc^.bs, npar, s) then
    Result := QUIRC_ERROR_DATA_ECC
  else
    Result := QUIRC_SUCCESS;
end;

(************************************************************************
 * Format value error correction
 *
 * Generator polynomial for GF(2^4) is x^4 + x + 1
 *)
const
  _FORMAT_MAX_ERROR = 3;
  _FORMAT_SYNDROMES = (_FORMAT_MAX_ERROR*2);
  _FORMAT_BITS = 15;

function format_syndromes(u: uint16_t;  s: puint8_t): boolean;
var
  i,j: integer;
begin
  Result := False;
  FillByte(s^, MAX_POLY, 0);
  for i:=0 to Pred(_FORMAT_SYNDROMES) do begin
    s[i] := 0;
    for j:=0 to Pred(_FORMAT_BITS) do
      if u and (1 shl j) <> 0 then
        s[i] := s[i] xor (gf16_exp[((i+1)*j) mod 15]);

    if s[i]<>0 then Result := True;
  end;
end;

function correct_format(var ret: uint16_t): quirc_decode_error_t;
var
  u: uint16_t;
  i: integer;
  s, sigma: array [0..Pred(MAX_POLY)] of uint8_t;
begin
  u := ret;
  
  (* Evaluate U (received codeword) at each of alpha_1 .. alpha_6
  * to get S_1 .. S_6 (but we index them from 0).
  *)

  if not format_syndromes(u,s) then begin
    Result := QUIRC_SUCCESS;
    Exit;
  end;

  berlekamp_massey(s, _FORMAT_SYNDROMES, @gf16, sigma);

  (* Now, find the roots of the polynomial *)
  for i:=0 to Pred(15) do
    if poly_eval(sigma, gf16_exp[15-i], @gf16) = 0 then
      u := u xor (1 shl i);

  if format_syndromes(u,s) then
    Result := QUIRC_ERROR_FORMAT_ECC
  else begin
    ret := u;
    Result := QUIRC_SUCCESS;
  end;
end;

(************************************************************************
 * Decoder algorithm
 *)
type
  datastream = record
    raw: array [0..Pred(QUIRC_MAX_PAYLOAD)] of uint8_t;
    data_bits: integer;
    ptr: integer;
    data: array [0..Pred(QUIRC_MAX_PAYLOAD)] of uint8_t;
  end;
  pdatastream = ^datastream;

function grid_bit(code: pquirc_code;  x: integer;  y: integer): uint8_t; inline;
var
  p: integer;
begin
  p := y*code^.size + x;
  Result := (code^.cell_bitmap[p shr 3] shr (p and 7)) and 1;
end;

function read_format(code: pquirc_code;  data: pquirc_data;  which: integer): quirc_decode_error_t; 
const
  xs: array [0..Pred(15)] of integer = (8,8,8,8,8,8,8,8,7,5,4,3,2,1,0);
  ys: array [0..Pred(15)] of integer = (0,1,2,3,4,5,7,8,8,8,8,8,8,8,8);
var
  i: integer;
  format, fdata: uint16_t;
begin
  // 15 bit format information
  format := 0;
  // which 0-TopLeft , 1-TopRight and BottomLeft
  if which<>0 then begin
    // on right side of Bottom-Left FP
    for i:=0 to Pred(7) do
      format := (format shl 1) or grid_bit(code, 8, code^.size-1-i);
    // under Top-Right FP
    for i:=0 to Pred(8) do
      format := (format shl 1) or grid_bit(code, code^.size-8+i, 8);
  end
  else begin
    // bits around Top-Left FP
    for i:=14 downto 0 do
      format := (format shl 1) or grid_bit(code, xs[i], ys[i]);
  end;

  format := format xor ($5412);
  Result := correct_format(format);
  if Result<>QUIRC_SUCCESS then Exit;

  fdata := format shr 10;
  data^.ecc_level := fdata shr 3; // 2 bits ecc level
  data^.mask := fdata and 7; // 3 bits mask
end;

function mask_bit(mask: integer;  i: integer;  j: integer): boolean;
begin
  case mask of
    0: Result := ((i+j) mod 2) = 0;
    1: Result := (i mod 2) = 0;
    2: Result := (j mod 3) = 0;
    3: Result := ((i+j) mod 3) = 0;
    4: Result := (((i div 2)+(j div 3)) mod 2) = 0;
    5: Result := ((i*j) mod 2+(i*j) mod 3) = 0;
    6: Result := (((i*j) mod 2+(i*j) mod 3) mod 2) = 0;
    7: Result := (((i*j) mod 3+(i+j) mod 2) mod 2) = 0;
    else Result := False;
  end;
end;

function reserved_cell(version: integer;  i: integer;  j: integer): boolean;
var
  ver: quirc_version_info;
  size, ai, aj, a, p: integer;
begin
  if version > High(quirc_version_db) then Exit(False); // not supported yet
  ver := quirc_version_db[version];
  size := version*4+17;
  (* Finder + format: top left *)
  if (i<9) and (j<9) then Exit(True);

  (* Finder + format: bottom left *)
  if (i+8>=size) and (j<9) then Exit(True);

  (* Finder + format: top right *)
  if (i<9) and (j+8>=size) then Exit(True);

  (* Exclude timing patterns *)
  if (i=6) or (j=6) then Exit(True);

  (* Exclude version info, if it exists. Version info sits adjacent to
   * the top-right and bottom-left finders in three rows, bounded by
   * the timing pattern.
   *)
  if version>=7 then begin
    if (i<6) and (j+11>=size) then Exit(True);

    if (i+11>=size) and (j<6) then Exit(True);
  end;

  (* Exclude alignment patterns *)
  ai := -1;
  aj := -1;
  a := 0;
  while (a < QUIRC_MAX_ALIGNMENT) and (ver.apat[a] <> 0) do begin
    p := ver.apat[a]; // position of alignment pattern
    if abs(p-i) < 3 then ai := a;
    if abs(p-j) < 3 then aj := a;
    Inc(a);
  end;

  if (ai>=0) and (aj>=0) then begin
    Dec(a);
    if (ai>0) and (ai<a) then Exit(True);
    if (aj>0) and (aj<a) then Exit(True);
    if (aj=a) and (ai=a) then Exit(True);
  end;

  Result := False;
end;

procedure read_bit(code: pquirc_code;  data: pquirc_data;  ds: pdatastream;  i: integer;  j: integer); 
var
  bitpos, bytepos, v: integer;
begin
  bitpos := ds^.data_bits and 7;
  bytepos := ds^.data_bits shr 3;

  v:=grid_bit(code, j, i);

  if mask_bit(data^.mask, i, j) then
    v := v xor 1;

  if v<>0 then
   ds^.raw[bytepos] := ds^.raw[bytepos] or ($80 shr bitpos);

  Inc(ds^.data_bits);
end;

procedure read_data(code: pquirc_code;  data: pquirc_data;  ds: pdatastream); 
var
  y, x, dir: integer;
begin
  // read data from bottom-right skipping reserved cells
  y := code^.size-1;
  x := code^.size-1;
  dir := -1;
  while x>0 do begin
    if x=6 then Dec(x);

    if not reserved_cell(data^.version,y,x) then
      read_bit(code,data,ds,y,x);
    if not reserved_cell(data^.version,y,x-1) then
      read_bit(code,data,ds,y,x-1);

    Inc(y, dir);
    if (y<0) or (y>=code^.size) then begin
      dir := -dir;
      Dec(x, 2);
      Inc(y, dir);
    end;
  end;
end;

function codestream_ecc(data: pquirc_data;  ds: pdatastream): quirc_decode_error_t; 
var
  ver: pquirc_version_info;
  lb_count, bc, ecc_offset, dst_offset, num_ec, i,j: integer;
  sb_ecc, ecc: pquirc_rs_params;
  lb_ecc: quirc_rs_params;
  dst: puint8_t;
begin
  ver := @quirc_version_db[data^.version];
  sb_ecc := @ver^.ecc[data^.ecc_level];
  lb_count := (ver^.data_bytes - sb_ecc^.bs * sb_ecc^.ns) div (sb_ecc^.bs + 1);
  bc := lb_count + sb_ecc^.ns;
  ecc_offset := sb_ecc^.dw * bc + lb_count;
  dst_offset := 0;
  
  lb_ecc := sb_ecc^;
  Inc(lb_ecc.dw);
  Inc(lb_ecc.bs);

  for i:=0 to Pred(bc) do begin
    dst := @ds^.data[dst_offset]; //
    if i < sb_ecc^.ns then
      ecc := sb_ecc
    else
      ecc := @lb_ecc;
    num_ec := ecc^.bs - ecc^.dw;

    for j:=0 to Pred(ecc^.dw) do
      dst[j] := ds^.raw[j*bc+i];
    for j:=0 to Pred(num_ec) do
      dst[ecc^.dw+j] := ds^.raw[ecc_offset+j*bc+i];

    Result := correct_block(dst, ecc);
    if Result <> QUIRC_SUCCESS then Exit;

    Inc(dst_offset, ecc^.dw);
  end;

  ds^.data_bits := dst_offset*8;
  Result := QUIRC_SUCCESS;
end;

function bits_remaining(ds: pdatastream): int; inline;
begin
   Result := ds^.data_bits - ds^.ptr;
end;

function take_bits(ds: pdatastream;  len: integer): int;
var
  b: uint8_t;
  bitpos: integer;
begin
  Result := 0;
  while (len<>0) and (ds^.ptr < ds^.data_bits) do begin
    b := ds^.data[ds^.ptr shr 3];
    bitpos := ds^.ptr and 7;

    Result := Result shl 1;
    if ((b shl bitpos) and $80) <> 0 then
      Result := Result or 1;

    Inc(ds^.ptr);
    Dec(len);
  end;
end;

function numeric_tuple(data: pquirc_data;  ds: pdatastream;  bits: integer;  digits: integer): integer; 
var
  tuple, i: integer;
begin
  if bits_remaining(ds)<bits then Exit(-1);

  tuple := take_bits(ds,bits);

  for i:=digits-1 downto 0 do begin
    data^.payload[data^.payload_len+i] := tuple mod 10 + Ord('0');
    tuple := tuple div 10;
  end;

  Inc(data^.payload_len, digits);
  Result := 0;
end;

function decode_numeric(data: pquirc_data;  ds: pdatastream): quirc_decode_error_t; 
var
  bits, count: integer;
begin
  bits := 14;
  
  if data^.version<10 then
    bits := 10
  else if data^.version<27 then
    bits := 12;

  count := take_bits(ds,bits);

  if data^.payload_len+count+1 > QUIRC_MAX_PAYLOAD then Exit(QUIRC_ERROR_DATA_OVERFLOW);

  while count>=3 do begin
    if numeric_tuple(data,ds,10,3) < 0 then Exit(QUIRC_ERROR_DATA_UNDERFLOW);
    Dec(count, 3);
  end;

  if count>=2 then begin
    if numeric_tuple(data,ds,7,2) < 0 then Exit(QUIRC_ERROR_DATA_UNDERFLOW);
    Dec(count, 2);
  end;

  if count<>0 then begin
    if numeric_tuple(data,ds,4,1) < 0 then Exit(QUIRC_ERROR_DATA_UNDERFLOW);
    Dec(count);
  end;

  Result := QUIRC_SUCCESS;
end;

function alpha_tuple(data: pquirc_data;  ds: pdatastream;  bits: integer;  digits: integer): integer; 
const
  alpha_map: PAnsiChar = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ $%*+-./:';
var
  tuple, i: integer;
begin
  if bits_remaining(ds) < bits then Exit(-1);

  tuple:= take_bits(ds,bits);

  for i:=0 to Pred(digits) do begin
    data^.payload[data^.payload_len+digits-i-1] := Ord(alpha_map[tuple mod 45]);
    tuple := tuple div 45;
  end;

  Inc(data^.payload_len, digits);
  Result := 0;
end;

function decode_alpha(data: pquirc_data;  ds: pdatastream): quirc_decode_error_t; 
var
  bits, count: integer;
begin
  bits:=13; 
  
  if data^.version<10 then
    bits := 9
  else if data^.version<27 then
    bits := 11;

  count := take_bits(ds,bits);

  if data^.payload_len+count+1 > QUIRC_MAX_PAYLOAD then Exit(QUIRC_ERROR_DATA_OVERFLOW);

  while count>=2 do begin
    if alpha_tuple(data,ds,11,2) < 0 then Exit(QUIRC_ERROR_DATA_UNDERFLOW);
    Dec(count, 2);
  end;

  if count<>0 then begin
    if alpha_tuple(data,ds,6,1) < 0 then Exit(QUIRC_ERROR_DATA_UNDERFLOW);
    Dec(count);
  end;

  Result := QUIRC_SUCCESS;
end;

function decode_byte(data: pquirc_data;  ds: pdatastream): quirc_decode_error_t; 
var
  bits, count, i: integer;
begin
  bits:=16; 
  
  if data^.version<10 then
    bits := 8;

  count := take_bits(ds,bits);

  if data^.payload_len+count+1 > QUIRC_MAX_PAYLOAD then Exit(QUIRC_ERROR_DATA_OVERFLOW);

  if bits_remaining(ds) < count*8 then Exit(QUIRC_ERROR_DATA_UNDERFLOW);

  for i:=0 to Pred(count) do begin
    data^.payload[data^.payload_len{++}] := take_bits(ds,8);
    Inc(data^.payload_len);
  end;

  Result := QUIRC_SUCCESS;
end;

function decode_kanji(data: pquirc_data;  ds: pdatastream): quirc_decode_error_t; 
var
  bits, count, i, d, msB, lsB, intermediate: integer;
  sjw: uint16_t;
begin
  bits:=12;

  if data^.version<10 then
    bits := 8
  else if data^.version<27 then
    bits := 10;

  count := take_bits(ds,bits);

  if data^.payload_len+count*2+1 > QUIRC_MAX_PAYLOAD then Exit(QUIRC_ERROR_DATA_OVERFLOW);

  if bits_remaining(ds) < count*13 then Exit(QUIRC_ERROR_DATA_UNDERFLOW);

  for i:=0 to Pred(count) do begin
    d := take_bits(ds,13);
    msB := d div $c0;
    lsB := d mod $c0;
    intermediate := (msB shl 8) or lsB;
    
    if intermediate+$8140 <= $9ffc then
      (* bytes are in the range 0x8140 to 0x9FFC *)
      sjw := intermediate+$8140
    else
      (* bytes are in the range 0xE040 to 0xEBBF *)
      sjw := intermediate+$c140;

    data^.payload[data^.payload_len{++}] := sjw shr 8;
    Inc(data^.payload_len);
    data^.payload[data^.payload_len{++}] := sjw and $ff;
    Inc(data^.payload_len);
  end;

  Result := QUIRC_SUCCESS;
end;

function decode_eci(data: pquirc_data;  ds: pdatastream): quirc_decode_error_t; 
begin
  if bits_remaining(ds)<8 then Exit(QUIRC_ERROR_DATA_UNDERFLOW);

  data^.eci := take_bits(ds,8);
  if (data^.eci and $c0)=$80 then begin
    if bits_remaining(ds) < 8 then Exit(QUIRC_ERROR_DATA_UNDERFLOW);
    data^.eci := (data^.eci shl 8) or take_bits(ds,8);
  end
  else if (data^.eci and $e0)=$c0 then begin
    if bits_remaining(ds) < 16 then Exit(QUIRC_ERROR_DATA_UNDERFLOW);
    data^.eci := (data^.eci shl 16) or take_bits(ds,16);
  end;

  Result:= QUIRC_SUCCESS;
end;

function decode_payload(data: pquirc_data;  ds: pdatastream): quirc_decode_error_t;
var
  t: int;
begin
  while bits_remaining(ds)>=4 do begin
    Result := QUIRC_SUCCESS;
    t := take_bits(ds,4);
    case t of
      QUIRC_DATA_TYPE_NUMERIC:
        Result := decode_numeric(data,ds);
      QUIRC_DATA_TYPE_ALPHA:
        Result := decode_alpha(data,ds);
      QUIRC_DATA_TYPE_BYTE:
        Result := decode_byte(data,ds);
      QUIRC_DATA_TYPE_KANJI:
        Result := decode_kanji(data,ds);
      7:
        Result := decode_eci(data,ds);
      else
        break; // goto done;
    end;

    if Result<>QUIRC_SUCCESS then Exit;

    if ((t and (t-1))=0) and (t > data^.data_type) then
      data^.data_type := t;
  end;

//done:
  (* Add nul terminator to all payloads *)
  if data^.payload_len >= sizeof(data^.payload) then
    Dec(data^.payload_len);
  data^.payload[data^.payload_len] := 0;

  Result:= QUIRC_SUCCESS;
end;

function quirc_decode(code: pquirc_code; out data: quirc_data): quirc_decode_error_t;
var
  ds: datastream;
begin
  if ((code^.size-17) mod 4)<>0 then Exit(QUIRC_ERROR_INVALID_GRID_SIZE);

  FillByte(data, sizeof(data), 0);
  FillByte(ds, sizeof(ds), 0);

  data.version := (code^.size-17) div 4;
  if (data.version<1) or (data.version>QUIRC_MAX_VERSION) then Exit(QUIRC_ERROR_INVALID_VERSION);

  (* Read format information -- try both locations *)
  Result := read_format(code, @data, 0);
  if Result <> QUIRC_SUCCESS then
    Result := read_format(code, @data, 1);
  if Result <> QUIRC_SUCCESS then
    Exit;

  read_data(code, @data, @ds); // datastream.raw
  Result := codestream_ecc(@data, @ds); // datastream.data
  if Result <> QUIRC_SUCCESS then Exit;

  Result := decode_payload(@data, @ds); // data.payload
end;

end.
