unit quircidentify;
(* quirc - QR-code recognition library
 * Copyright (C) 2010-2012 Daniel Beer <dlbeer@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)

(* C to Pascal translation by Ladislav Karrach in 2018,2020 (with help of c2pas32)
 *)

interface

uses quirc;

(* These functions are used to process images for QR-code recognition.
 * quirc_begin() must first be called to obtain access to a buffer into
 * which the input image should be placed. Optionally, the current
 * width and height may be returned.
 *
 * After filling the buffer, quirc_end() should be called to process
 * the image for QR-code recognition. The locations and content of each
 * code may be obtained using accessor functions described below.
 *)
function quirc_begin(q: pquirc;  w, h: pinteger): puint8_t;
procedure quirc_end(q: pquirc);

(* Extract the QR-code specified by the given index. *)
procedure quirc_extract(q: pquirc;  index: integer;  out code: quirc_code);


implementation

uses quircdecode {quirc_version_db};

{$DEFINE USE_OTSU} // use newer Otsu thresholding (2019)

(************************************************************************
 * Linear algebra routines
 *)

function line_intersect(p0, p1, q0, q1: pquirc_point; out r: quirc_point): boolean;
var
  a, b, c, d, e, f, det: integer;
begin
  (* (a, b) is perpendicular to line p *)
  a := -(p1^.y - p0^.y);
  b := p1^.x - p0^.x;

  (* (c, d) is perpendicular to line q *)
  c := -(q1^.y - q0^.y);
  d := q1^.x - q0^.x;

  (* e and f are dot products of the respective vectors with p and q *)
  e := a*p1^.x + b*p1^.y;
  f := c*q1^.x + d*q1^.y;

  (* Now we need to solve:
    *     [a b] [rx]   [e]
    *     [c d] [ry] = [f]
    *
    * We do this by inverting the matrix and applying it to (e, f):
    *       [ d -b] [e]   [rx]
    * 1/det [-c  a] [f] = [ry]
    *)
  det := (a*d)-(b*c);
  if det = 0 then
    Result := False
  else begin
    r.x := (d*e-b*f) div det;
    r.y := (-c*e+a*f) div det;
    Result := True;
  end;
end;

procedure perspective_setup(c: pdouble;  rect: pquirc_point;  w, h: double);
var
  x0, y0, x1, y1, x2, y2, x3, y3, wden, hden: double;
begin
  x0:=rect[0].x; 
  y0:=rect[0].y; 
  x1:=rect[1].x; 
  y1:=rect[1].y; 
  x2:=rect[2].x; 
  y2:=rect[2].y; 
  x3:=rect[3].x; 
  y3:=rect[3].y; 
  wden := w*(x2*y3-x3*y2+(x3-x2)*y1+x1*(y2-y3));
  hden := h*(x2*y3+x1*(y2-y3)-x3*y2+(x3-x2)*y1);
  if (wden=0) or (hden=0) then begin // my addition
    c[0] := 0;
    c[1] := 0;
    c[2] := x0;
    c[3] := 0;
    c[4] := 0;
    c[5] := y0;
    c[6] := 0;
    c[7] := 0;
  end
  else begin
    c[0] := (x1*(x2*y3-x3*y2)+x0*(-x2*y3+x3*y2+(x2-x3)*y1)+x1*(x3-x2)*y0) / wden;
    c[1] := -(x0*(x2*y3+x1*(y2-y3)-x2*y1)-x1*x3*y2+x2*x3*y1+(x1*x3-x2*x3)*y0) / hden;
    c[2] := x0;
    c[3] := (y0*(x1*(y3-y2)-x2*y3+x3*y2)+y1*(x2*y3-x3*y2)+x0*y1*(y2-y3)) / wden;
    c[4] := (x0*(y1*y3-y2*y3)+x1*y2*y3-x2*y1*y3+y0*(x3*y2-x1*y2+(x2-x3)*y1)) / hden;
    c[5] := y0;
    c[6] := (x1*(y3-y2)+x0*(y2-y3)+(x2-x3)*y1+(x3-x2)*y0) / wden;
    c[7] := (-x2*y3+x1*y3+x3*y2+x0*(y1-y2)-x3*y1+(x2-x1)*y0) / hden;
  end;
end;

function rint(d: double): integer; inline;
begin
  Result := Trunc(d + 0.5);
  //Result := Round(d);
end;

procedure perspective_map(c: pdouble;  u, v: double;  out ret: quirc_point);
var
  den, x, y: double;
begin
  den := c[6]*u+c[7]*v+1.0;
  x := (c[0]*u+c[1]*v+c[2]) / den;
  y := (c[3]*u+c[4]*v+c[5]) / den;
  ret.x := rint(x);
  ret.y := rint(y);
end;

procedure perspective_unmap(c: pdouble;  &in: pquirc_point; out u, v: double);
var
  x, y, den: double;
begin
  x := &in^.x;
  y := &in^.y;
  den := -c[0]*c[7]*y + c[1]*c[6]*y + (c[3]*c[7]-c[4]*c[6])*x + c[0]*c[4] - c[1]*c[3];
  if den = 0 then begin // my addition
    u := 0;
    v := 0;
  end
  else begin
    u := -(c[1]*(y-c[5]) - c[2]*c[7]*y + (c[5]*c[7]-c[4])*x + c[2]*c[4]) / den;
    v := (c[0]*(y-c[5]) - c[2]*c[6]*y + (c[5]*c[6]-c[3])*x + c[2]*c[3]) / den;
  end;
end;

(************************************************************************
 * Span-based floodfill routine
 *)
const
  FLOOD_FILL_MAX_DEPTH = 4096;

type
  span_func_t = procedure(user_data: pointer;  y, left, right: integer);

procedure flood_fill_seed(q: pquirc;  x, y: integer;  from, &to: integer;  func: span_func_t;  user_data: pointer;  depth: integer);
var
  left, right: integer;
  i: integer;
  row: pquirc_pixel_t;
begin
  left:=x; 
  right:=x; 
  
  row:=q^.pixels + y*q^.w;
  if depth>=FLOOD_FILL_MAX_DEPTH then Exit;

  while (left>0) and (row[left-1]=from) do
    Dec(left);

  while (right<q^.w-1) and (row[right+1]=from) do
    Inc(right);

  (* Fill the extent *)
  for i:=left to right do
    row[i]:= &to;

  if Assigned(func) then
    func(user_data,y,left,right);

  (* Seed new flood-fills *)
  if y>0 then begin
    row := q^.pixels+(y-1)*q^.w;
    for i:=left to right do
      if row[i]=from then
        flood_fill_seed(q, i, y-1, from, &to, func, user_data, depth+1);
  end;

  if y < q^.h-1 then begin
    row := q^.pixels+(y+1)*q^.w;
    for i:=left to right do
      if row[i]=from then
        flood_fill_seed(q, i, y+1, from, &to, func, user_data, depth+1);
  end;
end;

(************************************************************************
 * Adaptive thresholding
 *)
procedure threshold(q: pquirc);
const
  THRESHOLD_S_MIN = 1;
  THRESHOLD_S_DEN = 8;
  THRESHOLD_T = 5;
var
  x, y, avg_w, avg_u, threshold_s: integer;
  row: pquirc_pixel_t;
  w, u: integer;
begin
  avg_w := 0;
  avg_u := 0;
  threshold_s := q^.w div THRESHOLD_S_DEN;
  row := q^.pixels;
  (*
   * Ensure a sane, non-zero value for threshold_s.
   *
   * threshold_s can be zero if the image width is small. We need to avoid
   * SIGFPE as it will be used as divisor.
   *)
  if threshold_s < THRESHOLD_S_MIN then
    threshold_s := THRESHOLD_S_MIN;

  for y:=0 to Pred(q^.h) do begin
    FillByte(q^.row_average^, q^.w*sizeof(int), 0);
    for x:=0 to Pred(q^.w) do begin
      if (y and 1) <> 0 then begin
        w := x;
        u := q^.w-1-x;
      end
      else begin
        w := q^.w-1-x;
        u := x;
      end;
      avg_w := (avg_w*(threshold_s-1)) div threshold_s + row[w];
      avg_u := (avg_u*(threshold_s-1)) div threshold_s + row[u];
      q^.row_average[w] := q^.row_average[w] + avg_w;
      q^.row_average[u] := q^.row_average[u] + avg_u;
    end;

    for x:=0 to Pred(q^.w) do begin
      if row[x]<q^.row_average[x]*(100-THRESHOLD_T) div (200*threshold_s) then
        row[x] := QUIRC_PIXEL_BLACK
      else
        row[x] := QUIRC_PIXEL_WHITE;
    end;

    row := row + q^.w;
  end;
end;

(************************************************************************
 * Adaptive thresholding
 * (original threshold() replaced by otsu())
 *)
function otsu(q: pquirc): uint8_t;
const
  UINT8_MAX = 255;
var
  numPixels, length, sum, sumB, q1, q2: integer;
  i: DWord;
  histogram: array [0..UINT8_MAX] of DWord;
  ptr: puint8_t;
  max, m1, m2, variance: double;
begin
  numPixels := q^.w*q^.h;

  (* Calculate histogram*)
  FillByte(histogram, SizeOf(histogram), 0);
  ptr := q^.image;
  length := numPixels;
  while length > 0 do begin
    Inc(histogram[ptr^]);
    Inc(ptr);
    Dec(length);
  end;

  (* Calculate weighted sum of histogram values*)
  sum:=0;
  for i:=0 to UINT8_MAX do
    Inc(sum, i*histogram[i]);

  (* Compute threshold*)
  sumB:=0;
  q1:=0;
  max:=0;
  Result:=0;
  for i:=0 to UINT8_MAX do begin
    (* Weighted background*)
    Inc(q1, histogram[i]);
    if q1=0 then continue;

    (* Weighted foreground*)
    q2 := numPixels-q1;
    if q2=0 then break;

    Inc(sumB, i*histogram[i]);
    m1 := sumB/q1;
    m2 := (sum-sumB)/q2;
    variance := Sqr(m1-m2)*q1*q2;

    if variance >= max then begin
      Result := i;
      max := variance;
    end;
  end;
end;

procedure area_count(user_data: pointer;  y, left, right: integer);
begin
  Inc(pquirc_region(user_data)^.count, right-left+1);
end;

function region_code(q: pquirc;  x, y: integer): integer;
var
  pixel: integer;
  box: pquirc_region;
begin
  if (x<0) or (y<0) or (x>=q^.w) or (y>=q^.h) then Exit(-1);

  pixel := q^.pixels[y*q^.w+x];
  if pixel >= QUIRC_PIXEL_REGION then Exit(pixel);

  if pixel=QUIRC_PIXEL_WHITE then Exit(-1);

  if q^.num_regions>=QUIRC_MAX_REGIONS then Exit(-2); //Halt(-2);

  Result := q^.num_regions;
  box := @q^.regions[q^.num_regions{++}]; Inc(q^.num_regions);
  FillByte(box^, sizeof(box^), 0);

  box^.seed.x := x;
  box^.seed.y := y;
  box^.capstone := -1;
  flood_fill_seed(q, x, y, pixel, Result, @area_count, box, 0);
end;

type
  polygon_score_data = record
    ref: quirc_point;
    scores: array [0..3] of int;
    corners: pquirc_point;
  end;
  ppolygon_score_data = ^polygon_score_data;

procedure find_one_corner(user_data: pointer;  y, left, right: integer);
var
  psd: ppolygon_score_data;
  xs: array [0..1] of integer;
  i, dx, dy, d: integer;
begin
  psd := ppolygon_score_data(user_data);
  xs[0] := left;
  xs[1] := right;
  dy := y-psd^.ref.y;
  
  for i:=0 to 1 do begin
    dx := xs[i] - psd^.ref.x;
    d := dx*dx+dy*dy;
    if d > psd^.scores[0] then begin
      psd^.scores[0] := d;
      psd^.corners[0].x := xs[i];
      psd^.corners[0].y := y;
    end;
  end;
end;

procedure find_other_corners(user_data: pointer;  y, l, r: integer);
var
  psd: ppolygon_score_data;
  xs: array [0..1] of integer;
  i, j, up, right: integer;
  scores: array [0..3] of integer;
begin
  psd := ppolygon_score_data(user_data);
  xs[0] := l;
  xs[1] := r;

  for i:=0 to 1 do begin
    up := xs[i]*psd^.ref.x + y*psd^.ref.y;
    right := xs[i]*-psd^.ref.y + y*psd^.ref.x;
    scores[0] := up;
    scores[1] := right;
    scores[2] := -up;
    scores[3] := -right;
    
    for j:=0 to 3 do begin
      if scores[j] > psd^.scores[j] then begin
        psd^.scores[j] := scores[j];
        psd^.corners[j].x := xs[i];
        psd^.corners[j].y := y;
      end;
    end;
  end;
end;

procedure find_region_corners(q: pquirc;  rcode: integer;  ref, corners: pquirc_point);
var
  region: pquirc_region;
  psd: polygon_score_data;
  i: integer;
begin
  region :=  @q^.regions[rcode];
  FillByte(psd, sizeof(psd), 0);
  psd.corners := corners;

  Move(ref^, psd.ref, sizeof(psd.ref));
  psd.scores[0] := -1;

  flood_fill_seed(q, region^.seed.x, region^.seed.y, rcode, QUIRC_PIXEL_BLACK, @find_one_corner, @psd, 0);

  psd.ref.x := psd.corners[0].x-psd.ref.x;
  psd.ref.y := psd.corners[0].y-psd.ref.y;

  for i:=0 to Pred(4) do
    Move(region^.seed, psd.corners[i], sizeof(psd.corners[i]));

  i := region^.seed.x*psd.ref.x + region^.seed.y*psd.ref.y;
  psd.scores[0] := i;
  psd.scores[2] := -i;
  i := region^.seed.x*-psd.ref.y + region^.seed.y*psd.ref.x;
  psd.scores[1] := i;
  psd.scores[3] := -i;

  flood_fill_seed(q, region^.seed.x, region^.seed.y, QUIRC_PIXEL_BLACK, rcode, @find_other_corners, @psd, 0);
end;

procedure record_capstone(q: pquirc;  ring: integer;  stone: integer); 
var
  stone_reg, ring_reg: pquirc_region;
  capstone: pquirc_capstone;
  cs_index: integer;
begin
  stone_reg := @q^.regions[stone];
  ring_reg := @q^.regions[ring];

  if q^.num_capstones>=QUIRC_MAX_CAPSTONES then Exit;

  cs_index := q^.num_capstones;
  capstone := @q^.capstones[q^.num_capstones{++}]; Inc(q^.num_capstones);
  FillByte(capstone^, sizeof(capstone^), 0);

  capstone^.qr_grid := -1;
  capstone^.ring := ring;
  capstone^.stone := stone;
  stone_reg^.capstone := cs_index;
  ring_reg^.capstone := cs_index;

  (* Find the corners of the ring *)
  find_region_corners(q, ring, @stone_reg^.seed, capstone^.corners);

  (* Set up the perspective transform and find the center *)
  perspective_setup(capstone^.c, capstone^.corners, 7.0, 7.0);
  perspective_map(capstone^.c, 3.5, 3.5, capstone^.center);
end;

procedure test_capstone(q: pquirc;  x, y: integer;  pb: pinteger);
var
  ring_right, ring_left, stone, ratio: integer;
  stone_reg, ring_reg: pquirc_region;
begin
  ring_right := region_code(q, x-pb[4], y);
  stone := region_code(q, x-pb[4]-pb[3]-pb[2], y);
  ring_left := region_code(q, x-pb[4]-pb[3]-pb[2]-pb[1]-pb[0], y);
  
  if (ring_left<0) or (ring_right<0) or (stone<0) then Exit;

  (* Left and ring of ring should be connected *)
  if ring_left<>ring_right then Exit;

  (* Ring should be disconnected from stone *)
  if ring_left=stone then Exit;

  stone_reg := @q^.regions[stone];
  ring_reg := @q^.regions[ring_left];

  (* Already detected *)
  if (stone_reg^.capstone>=0) or (ring_reg^.capstone>=0) then Exit;

  (* Ratio should ideally be 37.5 *)
  ratio := (stone_reg^.count*100) div ring_reg^.count;
  if (ratio<10) or (ratio>70) then Exit;

  record_capstone(q, ring_left, stone);
end;



procedure finder_scan(q: pquirc;  y: integer);
const
  check: array [0..4] of integer = (1, 1, 3, 1, 1);
var
  row: pquirc_pixel_t;
  x, last_color, run_length, run_count: integer;
  pb: array [0..4] of integer;
  color: integer;
  avg, err, i: integer;
  ok: boolean;
begin
  row := q^.pixels + y*q^.w;
  
  last_color:=0; 
  run_length:=0; 
  run_count:=0; 
  
  FillByte(pb, sizeof(pb), 0);
  for x:=0 to Pred(q^.w) do begin
    color := Ord(row[x] <> 0);
    if (x>0) and (color<>last_color) then begin
      Move(pb[1], pb[0], sizeof(pb[0])*4);
      pb[4] := run_length;
      run_length := 0;
      Inc(run_count);
      if (color=0) and (run_count>=5) then begin
        ok := True;
        avg := (pb[0]+pb[1]+pb[3]+pb[4]) div 4;
        err := (avg*3) div 4;

        for i:=0 to 4 do
          if (pb[i]<check[i]*avg-err) or (pb[i]>check[i]*avg+err) then
            ok := False;

        if ok then
          test_capstone(q,x,y,pb);
      end;
    end;
    Inc(run_length);
    last_color := color;
  end;
end;

procedure find_alignment_pattern(q: pquirc;  index: integer);
const
  dx_map: array [0..3] of integer = (1,0,-1,0);
  dy_map: array [0..3] of integer = (0,-1,0,1);
var
  qr: pquirc_grid;
  c0, c2: pquirc_capstone;
  a, b, c: quirc_point;
  size_estimate, step_size, dir: integer;
  u, v: double;
  i, code: integer;
  reg: pquirc_region;
begin
  qr := @q^.grids[index];
  c0 := @q^.capstones[qr^.caps[0]];
  c2 := @q^.capstones[qr^.caps[2]];

  (* Grab our previous estimate of the alignment pattern corner *)
  Move(qr^.align, b, sizeof(b));

  (* Guess another two corners of the alignment pattern so that we
   * can estimate its size.
   *)
  perspective_unmap(c0^.c, @b, u, v);
  perspective_map(c0^.c, u, v+1, a);
  perspective_unmap(c2^.c, @b, u, v);
  perspective_map(c2^.c, u+1, v, c);

  size_estimate := Abs((a.x-b.x)*-(c.y-b.y) + (a.y-b.y)*(c.x-b.x));

  (* Spiral outwards from the estimate point until we find something
   * roughly the right size. Don't look too from the estimate
   * point.
   *)
  step_size:=1;
  dir:=0;
  while step_size*step_size < size_estimate*100 do begin
    for i:=0 to Pred(step_size) do begin
      code := region_code(q, b.x, b.y);
      if code>=0 then begin
        reg := @q^.regions[code];
        if (reg^.count>=size_estimate div 2) and (reg^.count<=size_estimate*2) then begin
          qr^.align_region := code;
          Exit;
        end;
      end;
      Inc(b.x, dx_map[dir]);
      Inc(b.y, dy_map[dir]);
    end;

    dir := (dir+1) mod 4;
    if (dir and 1) = 0 then
      Inc(step_size);
  end;
end;

procedure find_leftmost_to_line(user_data: pointer;  y, left, right: integer);
var
  psd: ppolygon_score_data;
  xs: array [0..1] of integer;
  i, d: integer;
begin
  psd := ppolygon_score_data(user_data);
  xs[0] := left;
  xs[1] := right;
  
  for i:=0 to 1 do begin
    d := -psd^.ref.y*xs[i] + psd^.ref.x*y;
    if d < psd^.scores[0] then begin
      psd^.scores[0] := d;
      psd^.corners[0].x := xs[i];
      psd^.corners[0].y := y;
    end;
  end;
end;

(* Do a Bresenham scan from one point to another and count the number
 * of black/white transitions.
 *)
function timing_scan(q: pquirc;  p0, p1: pquirc_point): integer;
var
  n, d, x, y: integer;
  dom, nondom: pinteger;
  dom_step, nondom_step: integer;
  a, i, run_length, swap, pixel: integer;
begin
  n := p1^.x-p0^.x;
  d := p1^.y-p0^.y;

  if (p0^.x<0) or (p0^.y<0) or (p0^.x>=q^.w) or (p0^.y>=q^.h) then Exit(-1);
  if (p1^.x<0) or (p1^.y<0) or (p1^.x>=q^.w) or (p1^.y>=q^.h) then Exit(-1);

  if abs(n)>abs(d) then begin
    swap:=n; 
    n:= d; 
    d:= swap; 
    dom := @x;
    nondom := @y;
  end
  else begin
    dom := @y;
    nondom := @x;
  end;

  if n<0 then begin
    n := -n;
    nondom_step := -1;
  end
  else
    nondom_step := 1;

  if d<0 then begin
    d := -d;
    dom_step := -1;
  end
  else
    dom_step := 1;

  a:=0;
  run_length:=0;
  Result:=0;

  x := p0^.x;
  y := p0^.y;
  for i:=0 to d do begin
    if (y<0) or (y>=q^.h) or (x<0) or (x>=q^.w) then break;

    pixel := q^.pixels[x+y*q^.w];
    if pixel<>0 then begin
      if run_length>=2 then
        Inc(Result);
        run_length := 0;
    end
    else
      Inc(run_length);
    Inc(a, n);
    Inc(dom^, dom_step);
    if a>=d then begin
      Inc(nondom^, nondom_step);
      Dec(a, d);
    end;
  end;
end;

(* Try the measure the timing pattern for a given QR code. This does
 * not require the global perspective to have been set up, but it
 * does require that the capstone corners have been set to their
 * canonical rotation.
 *
 * For each capstone, we find a point in the middle of the ring band
 * which is nearest the centre of the code. Using these points, we do
 * a horizontal and a vertical timing scan.
 *)
function measure_timing_pattern(q: pquirc;  index: integer): boolean;
const
  us: array [0..2] of double = (6.5, 6.5, 0.5);
  vs: array [0..2] of double = (0.5, 6.5, 6.5);
var
  qr: pquirc_grid;
  cap: pquirc_capstone;
  i, scan, ver, size: integer;
begin
  qr := @q^.grids[index];

  for i:=0 to 2 do begin
    cap := @q^.capstones[qr^.caps[i]];
    perspective_map(cap^.c, us[i], vs[i], qr^.tpep[i]);
  end;

  qr^.hscan := timing_scan(q, @qr^.tpep[1], @qr^.tpep[2]);
  qr^.vscan := timing_scan(q, @qr^.tpep[1], @qr^.tpep[0]);

  scan := qr^.hscan;
  if qr^.vscan>scan then
    scan := qr^.vscan;

  (* If neither scan worked, we can't go any further. *)
  if scan<0 then Exit(False);

  (* Choose the nearest allowable grid size *)
  size := scan*2+13;
  ver := (size-15) div 4;
  qr^.grid_size := ver*4+17;

  Result := True;
end;

(* Read a cell from a grid using the currently set perspective
 * transform. Returns +/- 1 for black/white, 0 for cells which are
 * out of image bounds.
 *)
function read_cell(q: pquirc;  index, x, y: integer): integer;
var
  qr: pquirc_grid;
  p: quirc_point;
begin
  qr := @q^.grids[index];
  
  perspective_map(qr^.c, x+0.5, y+0.5, p);

  if (p.y<0) or (p.y>=q^.h) or (p.x<0) or (p.x>=q^.w) then
    Result := 0
  else if q^.pixels[p.y*q^.w+p.x] <> 0 then
    Result := +1
  else
    Result := -1;
end;

function fitness_cell(q: pquirc;  index, x, y: integer): integer;
const
  offsets: array [0..2] of double = (0.3,0.5,0.7);
var
  qr: pquirc_grid;
  p: quirc_point;
  u, v: integer;
begin
  qr := @q^.grids[index];
  Result := 0;
  
  for v:=0 to 2 do
    for u:=0 to 2 do begin
      perspective_map(qr^.c, x+offsets[u], y+offsets[v], p);

      if (p.y<0) or (p.y>=q^.h) or (p.x<0) or (p.x>=q^.w) then continue;

      if q^.pixels[p.y*q^.w+p.x]<>0 then
        Inc(Result)
      else
        Dec(Result);
  end;
end;

function fitness_ring(q: pquirc;  index, cx, cy, radius: integer): integer;
var
  i: integer;
begin
  Result := 0;
  for i:=0 to Pred(radius*2) do begin
    Inc(Result, fitness_cell(q, index, cx-radius+i, cy-radius));
    Inc(Result, fitness_cell(q, index, cx-radius  , cy+radius-i));
    Inc(Result, fitness_cell(q, index, cx+radius  , cy-radius+i));
    Inc(Result, fitness_cell(q, index, cx+radius-i, cy+radius));
  end;
end;

function fitness_apat(q: pquirc;  index, cx, cy: integer): integer;
begin
 Result := fitness_cell(q,index,cx,cy) -
           fitness_ring(q,index,cx,cy,1) +
           fitness_ring(q,index,cx,cy,2);
end;

function fitness_capstone(q: pquirc;  index, x, y: integer): integer;
begin
  Inc(x, 3);
  Inc(y, 3);
  Result := fitness_cell(q,index,x,y) +
            fitness_ring(q,index,x,y,1) -
            fitness_ring(q,index,x,y,2) +
            fitness_ring(q,index,x,y,3);
end;

(* Compute a fitness score for the currently configured perspective
 * transform, using the features we expect to find by scanning the
 * grid.
 *)
function fitness_all(q: pquirc;  index: integer): integer;
var
  qr: pquirc_grid;
  info: pquirc_version_info;
  version, i, j, ap_count, expect: integer;
begin
  qr := @q^.grids[index];
  version := (qr^.grid_size-17) div 4;

  Result := 0;
  
  (* Check the timing pattern *)  
  for i:=0 to Pred(qr^.grid_size-14) do begin
    if (i and 1) <> 0 then
      expect := 1
    else
      expect := -1;
    Inc(Result, fitness_cell(q, index, i+7, 6)*expect);
    Inc(Result, fitness_cell(q, index, 6, i+7)*expect);
  end;

  (* Check capstones *)
  Inc(Result, fitness_capstone(q, index, 0, 0));
  Inc(Result, fitness_capstone(q, index, qr^.grid_size-7, 0));
  Inc(Result, fitness_capstone(q, index, 0, qr^.grid_size-7));

  if (version<0) or (version>QUIRC_MAX_VERSION) then Exit;
  info := @quirc_version_db[version];

  (* Check alignment patterns *)
  ap_count := 0;
  while (ap_count<QUIRC_MAX_ALIGNMENT) and (info^.apat[ap_count]<>0) do
    Inc(ap_count);

  for i:=1 to ap_count-2 do begin
    Inc(Result, fitness_apat(q, index, 6, info^.apat[i]));
    Inc(Result, fitness_apat(q, index, info^.apat[i], 6));
  end;

  for i:=1 to ap_count-1 do
    for j:=1 to ap_count-1 do
      Inc(Result, fitness_apat(q, index, info^.apat[i], info^.apat[j]));
end;

procedure jiggle_perspective(q: pquirc;  index: integer); 
var
  qr: pquirc_grid;
  best, pass, test: integer;
  adjustments: array [0..7] of double;
  i, j: integer;
  old, step, new: double;
begin
  qr := @q^.grids[index];
  best := fitness_all(q, index);

  for i:=0 to 7 do
    adjustments[i] := qr^.c[i]*0.02;

  for pass:=0 to 4 do begin
    for i:=0 to 15 do begin
      j := i shr 1;
      old := qr^.c[j];
      step := adjustments[j];
      
      if (i and 1)<>0 then
        new := old+step
      else
        new := old-step;

      qr^.c[j] := new;
      test := fitness_all(q,index);

      if test>best then
        best := test
      else
        qr^.c[j] := old;
    end;

    for i:=0 to 7 do
      adjustments[i] := adjustments[i] * 0.5;
  end;
end;

(* Once the capstones are in place and an alignment point has been
 * chosen, we call this function to set up a grid-reading perspective
 * transform.
 *)
procedure setup_qr_perspective(q: pquirc;  index: integer);
var
  qr: pquirc_grid;
  rect: array[0..3] of quirc_point;
begin
  qr := @q^.grids[index];

  (* Set up the perspective map for reading the grid *)
  Move(q^.capstones[qr^.caps[1]].corners[0], rect[0], sizeof(rect[0]));
  Move(q^.capstones[qr^.caps[2]].corners[0], rect[1], sizeof(rect[0]));
  Move(qr^.align, rect[2], sizeof(rect[0]));
  Move(q^.capstones[qr^.caps[0]].corners[0], rect[3], sizeof(rect[0]));

  perspective_setup(qr^.c, rect, qr^.grid_size-7, qr^.grid_size-7);

  jiggle_perspective(q, index);
end;

(* Rotate the capstone with so that corner 0 is the leftmost with respect
 * to the given reference line.
 *)
procedure rotate_capstone(cap: pquirc_capstone;  h0, hd: pquirc_point);
var
  copy: array[0..3] of quirc_point;
  p: pquirc_point;
  j, best, best_score, score: integer;
begin
  best := 0;
  best_score := MAXINT;
  for j:=0 to 3 do begin
    p := @cap^.corners[j];
    score := (p^.x-h0^.x)*-hd^.y + (p^.y-h0^.y)*hd^.x;
    if (j=0) or (score<best_score) then begin
      best := j;
      best_score := score;
    end;
  end;

  (* Rotate the capstone *)
  for j:=0 to 3 do
    Move(cap^.corners[(j+best) mod 4], copy[j], sizeof(copy[j]));

  Move(copy, cap^.corners, sizeof(cap^.corners));
  perspective_setup(cap^.c, cap^.corners, 7.0, 7.0);
end;

(* Start from some point inside the alignment pattern *)
procedure record_qr_grid(q: pquirc;  a, b, c: integer);
var
  h0, hd: quirc_point;
  i, qr_index, swap: integer;
  qr: pquirc_grid;
  cap: pquirc_capstone;
  psd: polygon_score_data;
  reg: pquirc_region;
label
  fail;
begin
  if q^.num_grids >= QUIRC_MAX_GRIDS then Exit;

  (* Construct the hypotenuse line from A to C. B should be to
   * the left of this line.
   *)
  Move(q^.capstones[a].center, h0, sizeof(h0));
  hd.x := q^.capstones[c].center.x - q^.capstones[a].center.x;
  hd.y := q^.capstones[c].center.y - q^.capstones[a].center.y;

  (* Make sure A-B-C is clockwise *)
  if (q^.capstones[b].center.x-h0.x)*-hd.y + (q^.capstones[b].center.y-h0.y)*hd.x > 0 then begin
    swap := a;
    a := c;
    c := swap;
    hd.x := -hd.x;
    hd.y := -hd.y;
  end;

  (* Record the grid and its components *)
  qr_index := q^.num_grids;
  qr := @q^.grids[q^.num_grids{++}]; Inc(q^.num_grids);

  FillByte(qr^, sizeof(qr^), 0);
  qr^.caps[0] := a;
  qr^.caps[1] := b;
  qr^.caps[2] := c;
  qr^.align_region := -1;

  (* Rotate each capstone so that corner 0 is top-left with respect
   * to the grid.
   *)
  for i:=0 to 2 do begin
    cap := @q^.capstones[qr^.caps[i]];
    rotate_capstone(cap, @h0, @hd);
    cap^.qr_grid := qr_index;
  end;

  (* Check the timing pattern. This doesn't require a perspective
   * transform.
   *)
  if not measure_timing_pattern(q, qr_index) then
    goto fail;

  (* Make an estimate based for the alignment pattern based on extending
   * lines from capstones A and C.
   *)
  if not line_intersect(@q^.capstones[a].corners[0],
                        @q^.capstones[a].corners[1],
                        @q^.capstones[c].corners[0],
                        @q^.capstones[c].corners[3],
                        qr^.align) then
    goto fail;

  (* On V2+ grids, we should use the alignment pattern. *)
  if qr^.grid_size>21 then begin
    (* Try to find the actual location of the alignment pattern. *)
    find_alignment_pattern(q, qr_index);

    (* Find the point of the alignment pattern closest to the
     * top-left of the QR grid.
     *)
    if qr^.align_region >= 0 then begin
      reg := @q^.regions[qr^.align_region];

      (* Start from some point inside the alignment pattern *)
      Move(reg^.seed, qr^.align, sizeof(qr^.align));

      Move(hd, psd.ref, sizeof(psd.ref));
      psd.corners := @qr^.align;
      psd.scores[0] := -hd.y*qr^.align.x + hd.x*qr^.align.y;

      flood_fill_seed(q, reg^.seed.x, reg^.seed.y, qr^.align_region, QUIRC_PIXEL_BLACK, nil, nil, 0);
      flood_fill_seed(q, reg^.seed.x, reg^.seed.y, QUIRC_PIXEL_BLACK, qr^.align_region, @find_leftmost_to_line, @psd, 0);
    end;
  end;

  setup_qr_perspective(q, qr_index);
  Exit;

  (* We've been unable to complete setup for this grid. Undo what we've
   * recorded and pretend it never happened.
   *)
fail:
  for i:=0 to 2 do
    q^.capstones[qr^.caps[i]].qr_grid := -1;
  Dec(q^.num_grids);
end;

type
  neighbour = record
    index: integer;
    distance: double;
  end;
  pneighbour = ^neighbour;

  neighbour_list = record
    n: array [0..Pred(QUIRC_MAX_CAPSTONES)] of neighbour;
    count: integer;
  end;
  pneighbour_list = ^neighbour_list;

procedure test_neighbours(q: pquirc;  i: integer;  hlist, vlist: pneighbour_list);
var
  hn, vn: pneighbour;
  j, k: integer;
  best_score, score: double;
  best_h, best_v: integer;
begin
  best_score := 0.0;
  best_h := -1;
  best_v := -1;

  (* Test each possible grouping *)
  for j:=0 to Pred(hlist^.count) do
    for k:=0 to Pred(vlist^.count) do begin
      hn := @hlist^.n[j];
      vn := @vlist^.n[k];
      score := Abs(1 - hn^.distance / vn^.distance);

      if score>2.5 then continue;

      if (best_h<0) or (score<best_score) then begin
        best_h := hn^.index;
        best_v := vn^.index;
        best_score := score;
      end;
  end;

  if (best_h<0) or (best_v<0) then Exit;

  record_qr_grid(q, best_h, i, best_v);
end;

procedure test_grouping(q: pquirc;  i: integer); 
var
  c1, c2: pquirc_capstone;
  hlist, vlist: neighbour_list;
  j: integer;
  u, v: double;
  n: pneighbour;
begin
  c1 := @q^.capstones[i];
  if c1^.qr_grid >= 0 then Exit;

  hlist.count := 0;
  vlist.count := 0;
  (* Look for potential neighbours by examining the relative gradients
   * from this capstone to others.
   *)
  for j:=0 to Pred(q^.num_capstones) do begin
    c2 := @q^.capstones[j];

    if (i=j) or (c2^.qr_grid>=0) then continue;

    perspective_unmap(c1^.c, @c2^.center, u, v);

    u := Abs(u-3.5);
    v := Abs(v-3.5);

    if u < 0.2*v then begin
      n := @hlist.n[hlist.count{++}]; Inc(hlist.count);
      n^.index := j;
      n^.distance := v;
    end;

    if v < 0.2*u then begin
      n := @vlist.n[vlist.count{++}]; Inc(vlist.count);
      n^.index := j;
      n^.distance := u;
    end;
  end;

  if (hlist.count=0) or (vlist.count=0) then Exit;

  test_neighbours(q, i, @hlist, @vlist);
end;

{$IFDEF USE_OTSU}
procedure pixels_setup(q: pquirc; threshold: uint8_t);
var
  source: puint8_t;
  dest: pquirc_pixel_t;
  length: integer;
begin
  if sizeof(q^.image^) = sizeof(q^.pixels^) then
    q^.pixels := pquirc_pixel_t(q^.image);

  source := q^.image;
  dest := q^.pixels;
  for length := q^.w*q^.h-1 downto 0 do begin
    if source^ < threshold then
      dest^ := QUIRC_PIXEL_BLACK
    else
      dest^ := QUIRC_PIXEL_WHITE;
    Inc(source);
    Inc(dest);
  end;
end;
{$ELSE}
procedure pixels_setup(q: pquirc); 
var
  x, y: integer;
begin
  if sizeof(q^.image^) = sizeof(q^.pixels^) then
    q^.pixels := pquirc_pixel_t(q^.image)
  else begin
    for y:=0 to Pred(q^.h) do
      for x:=0 to Pred(q^.w) do
        q^.pixels[y*q^.w+x] := q^.image[y*q^.w+x];
  end;
end;
{$ENDIF}

function quirc_begin(q: pquirc;  w, h: pinteger): puint8_t;
begin
  q^.num_regions := QUIRC_PIXEL_REGION;
  q^.num_capstones := 0;
  q^.num_grids := 0;
  if w<>nil then w^ := q^.w;
  if h<>nil then h^ := q^.h;
  Result := q^.image;
end;

procedure quirc_end(q: pquirc); 
var
  i: integer;
begin
{$IFDEF USE_OTSU}
  pixels_setup(q, otsu(q));
{$ELSE}
  pixels_setup(q);
  threshold(q);
{$ENDIF}

  for i:=0 to Pred(q^.h) do
    finder_scan(q, i);  // q^.capstones

  for i:=0 to Pred(q^.num_capstones) do
    test_grouping(q, i);
end;

procedure quirc_extract(q: pquirc;  index: integer;  out code: quirc_code);
var
  qr: pquirc_grid;
  y, x, i: integer;
begin
  qr := @q^.grids[index];
  
  if (index<0) or (index>q^.num_grids) then Exit;

  FillByte(code, sizeof(code), 0);

  perspective_map(qr^.c, 0.0, 0.0, code.corners[0]);
  perspective_map(qr^.c, qr^.grid_size, 0.0, code.corners[1]);
  perspective_map(qr^.c, qr^.grid_size, qr^.grid_size, code.corners[2]);
  perspective_map(qr^.c, 0.0, qr^.grid_size, code.corners[3]);

  code.size := qr^.grid_size;
  if code.size > 17 + 4*QUIRC_MAX_VERSION then Exit; // my addition

  i:=0;
  for y:=0 to Pred(qr^.grid_size) do
    for x:=0 to Pred(qr^.grid_size) do begin
      if read_cell(q, index, x, y) > 0 then
        code.cell_bitmap[i shr 3] := code.cell_bitmap[i shr 3] or (1 shl (i and 7));
      Inc(i);
    end;
end;

end.
